@@ -2,17 +2,18 @@
 #include <Python.h>
 #include "structmember.h"
 
+#define NPY_NO_DEPRECATED_API NPY_API_VERSION
 #define _MULTIARRAYMODULE
-#define NPY_NO_PREFIX
 #include "numpy/arrayobject.h"
 #include "numpy/arrayscalars.h"
 
 #include "npy_config.h"
 
-#include "numpy/npy_3kcompat.h"
+#include "npy_pycompat.h"
 
 #include "buffer.h"
 #include "numpyos.h"
+#include "arrayobject.h"
 
 /*************************************************************************
  ****************   Implement Buffer Protocol ****************************
@@ -20,6 +21,7 @@
 
 /* removed multiple segment interface */
 
+#if !defined(NPY_PY3K)
 static Py_ssize_t
 array_getsegcount(PyArrayObject *self, Py_ssize_t *lenp)
 {
@@ -44,7 +46,7 @@ array_getreadbuf(PyArrayObject *self, Py_ssize_t segment, void **ptrptr)
         return -1;
     }
     if (PyArray_ISONESEGMENT(self)) {
-        *ptrptr = self->data;
+        *ptrptr = PyArray_DATA(self);
         return PyArray_NBYTES(self);
     }
     PyErr_SetString(PyExc_ValueError, "array is not a single segment");
@@ -56,14 +58,10 @@ array_getreadbuf(PyArrayObject *self, Py_ssize_t segment, void **ptrptr)
 static Py_ssize_t
 array_getwritebuf(PyArrayObject *self, Py_ssize_t segment, void **ptrptr)
 {
-    if (PyArray_CHKFLAGS(self, WRITEABLE)) {
-        return array_getreadbuf(self, segment, (void **) ptrptr);
-    }
-    else {
-        PyErr_SetString(PyExc_ValueError, "array cannot be "
-                        "accessed as a writeable buffer");
+    if (PyArray_FailUnlessWriteable(self, "buffer source array") < 0) {
         return -1;
     }
+    return array_getreadbuf(self, segment, (void **) ptrptr);
 }
 
 static Py_ssize_t
@@ -71,6 +69,7 @@ array_getcharbuf(PyArrayObject *self, Py_ssize_t segment, constchar **ptrptr)
 {
     return array_getreadbuf(self, segment, (void **) ptrptr);
 }
+#endif /* !defined(NPY_PY3K) */
 
 
 /*************************************************************************
@@ -153,7 +152,7 @@ _is_natively_aligned_at(PyArray_Descr *descr,
 {
     int k;
 
-    if ((Py_ssize_t)(arr->data) % descr->alignment != 0) {
+    if ((Py_ssize_t)(PyArray_DATA(arr)) % descr->alignment != 0) {
         return 0;
     }
 
@@ -165,9 +164,9 @@ _is_natively_aligned_at(PyArray_Descr *descr,
         return 0;
     }
 
-    for (k = 0; k < arr->nd; ++k) {
-        if (arr->dimensions[k] > 1) {
-            if (arr->strides[k] % descr->alignment != 0) {
+    for (k = 0; k < PyArray_NDIM(arr); ++k) {
+        if (PyArray_DIM(arr, k) > 1) {
+            if (PyArray_STRIDE(arr, k) % descr->alignment != 0) {
                 return 0;
             }
         }
@@ -193,19 +192,27 @@ _buffer_format_string(PyArray_Descr *descr, _tmp_string_t *str,
     }
 
     if (descr->subarray) {
-        PyObject *item;
+        PyObject *item, *subarray_tuple;
         Py_ssize_t total_count = 1;
         Py_ssize_t dim_size;
         char buf[128];
         int old_offset;
         int ret;
 
+        if (PyTuple_Check(descr->subarray->shape)) {
+            subarray_tuple = descr->subarray->shape;
+            Py_INCREF(subarray_tuple);
+        }
+        else {
+            subarray_tuple = Py_BuildValue("(O)", descr->subarray->shape);
+        }
+
         _append_char(str, '(');
-        for (k = 0; k < PyTuple_GET_SIZE(descr->subarray->shape); ++k) {
+        for (k = 0; k < PyTuple_GET_SIZE(subarray_tuple); ++k) {
             if (k > 0) {
                 _append_char(str, ',');
             }
-            item = PyTuple_GET_ITEM(descr->subarray->shape, k);
+            item = PyTuple_GET_ITEM(subarray_tuple, k);
             dim_size = PyNumber_AsSsize_t(item, NULL);
 
             PyOS_snprintf(buf, sizeof(buf), "%ld", (long)dim_size);
@@ -213,6 +220,9 @@ _buffer_format_string(PyArray_Descr *descr, _tmp_string_t *str,
             total_count *= dim_size;
         }
         _append_char(str, ')');
+
+        Py_DECREF(subarray_tuple);
+
         old_offset = *offset;
         ret = _buffer_format_string(descr->subarray->base, str, arr, offset,
                                     active_byteorder);
@@ -220,6 +230,8 @@ _buffer_format_string(PyArray_Descr *descr, _tmp_string_t *str,
         return ret;
     }
     else if (PyDataType_HASFIELDS(descr)) {
+        int base_offset = *offset;
+
         _append_str(str, "T{");
         for (k = 0; k < PyTuple_GET_SIZE(descr->names); ++k) {
             PyObject *name, *item, *offset_obj, *tmp;
@@ -233,14 +245,20 @@ _buffer_format_string(PyArray_Descr *descr, _tmp_string_t *str,
 
             child = (PyArray_Descr*)PyTuple_GetItem(item, 0);
             offset_obj = PyTuple_GetItem(item, 1);
-            new_offset = PyInt_AsLong(offset_obj);
+            new_offset = base_offset + PyInt_AsLong(offset_obj);
 
             /* Insert padding manually */
+            if (*offset > new_offset) {
+                PyErr_SetString(PyExc_RuntimeError,
+                                "This should never happen: Invalid offset in "
+                                "buffer format string generation. Please "
+                                "report a bug to the Numpy developers.");
+                return -1;
+            }
             while (*offset < new_offset) {
                 _append_char(str, 'x');
                 ++*offset;
             }
-            *offset += child->elsize;
 
             /* Insert child item */
             _buffer_format_string(child, str, arr, offset,
@@ -287,6 +305,8 @@ _buffer_format_string(PyArray_Descr *descr, _tmp_string_t *str,
             descr->type_num == NPY_ULONGLONG);
 #endif
 
+        *offset += descr->elsize;
+
         if (descr->byteorder == '=' &&
                 _is_natively_aligned_at(descr, arr, *offset)) {
             /* Prefer native types, to cater for Cython */
@@ -314,7 +334,7 @@ _buffer_format_string(PyArray_Descr *descr, _tmp_string_t *str,
 
             if (is_native_only_type) {
                 /* It's not possible to express native-only data types
-                   in non-native byte orders */
+                   in non-native npy_byte orders */
                 PyErr_Format(PyExc_ValueError,
                              "cannot expose native-only dtype '%c' in "
                              "non-native byte order '%c' via buffer interface",
@@ -348,6 +368,7 @@ _buffer_format_string(PyArray_Descr *descr, _tmp_string_t *str,
             break;
         case NPY_LONGLONG:     if (_append_char(str, 'q')) return -1; break;
         case NPY_ULONGLONG:    if (_append_char(str, 'Q')) return -1; break;
+        case NPY_HALF:         if (_append_char(str, 'e')) return -1; break;
         case NPY_FLOAT:        if (_append_char(str, 'f')) return -1; break;
         case NPY_DOUBLE:       if (_append_char(str, 'd')) return -1; break;
         case NPY_LONGDOUBLE:   if (_append_char(str, 'g')) return -1; break;
@@ -583,12 +604,12 @@ array_getbuffer(PyObject *obj, Py_buffer *view, int flags)
 
     /* Check whether we can provide the wanted properties */
     if ((flags & PyBUF_C_CONTIGUOUS) == PyBUF_C_CONTIGUOUS &&
-        !PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)) {
+        !PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)) {
         PyErr_SetString(PyExc_ValueError, "ndarray is not C-contiguous");
         goto fail;
     }
     if ((flags & PyBUF_F_CONTIGUOUS) == PyBUF_F_CONTIGUOUS &&
-        !PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)) {
+        !PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)) {
         PyErr_SetString(PyExc_ValueError, "ndarray is not Fortran contiguous");
         goto fail;
     }
@@ -599,15 +620,26 @@ array_getbuffer(PyObject *obj, Py_buffer *view, int flags)
     }
     if ((flags & PyBUF_STRIDES) != PyBUF_STRIDES &&
         (flags & PyBUF_ND) == PyBUF_ND &&
-        !PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)) {
+        !PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)) {
         /* Non-strided N-dim buffers must be C-contiguous */
         PyErr_SetString(PyExc_ValueError, "ndarray is not C-contiguous");
         goto fail;
     }
-    if ((flags & PyBUF_WRITEABLE) == PyBUF_WRITEABLE &&
-        !PyArray_ISWRITEABLE(self)) {
-        PyErr_SetString(PyExc_ValueError, "ndarray is not writeable");
-        goto fail;
+    if ((flags & PyBUF_WRITEABLE) == PyBUF_WRITEABLE) {
+        if (PyArray_FailUnlessWriteable(self, "buffer source array") < 0) {
+            goto fail;
+        }
+    }
+    /*
+     * If a read-only buffer is requested on a read-write array, we return a
+     * read-write buffer, which is dubious behavior. But that's why this call
+     * is guarded by PyArray_ISWRITEABLE rather than (flags &
+     * PyBUF_WRITEABLE).
+     */
+    if (PyArray_ISWRITEABLE(self)) {
+        if (array_might_be_written(self) < 0) {
+            goto fail;
+        }
     }
 
     if (view == NULL) {
@@ -737,7 +769,7 @@ _descriptor_from_pep3118_format(char *s)
     PyObject *_numpy_internal;
 
     if (s == NULL) {
-        return PyArray_DescrNewFromType(PyArray_BYTE);
+        return PyArray_DescrNewFromType(NPY_BYTE);
     }
 
     /* Strip whitespace, except from field names */
@@ -768,7 +800,7 @@ _descriptor_from_pep3118_format(char *s)
         Py_DECREF(str);
         return NULL;
     }
-    descr = (PyArray_Descr*)PyObject_CallMethod(
+    descr = PyObject_CallMethod(
         _numpy_internal, "_dtype_from_pep3118", "O", str);
     Py_DECREF(str);
     Py_DECREF(_numpy_internal);
